#include "include/TensorVolume.h"
#include <string.h>
#include <libgen.h>
#include <iostream>


using namespace std;


static inline int is_big_endian()						//Determines if we are on a big endian or little endian platform
{
    union 
	{
        uint32_t i;
        char c[4];
    } bint = {0x01020304};

    return bint.c[0] == 1; 
}



inline float ReverseFloat(const float inFloat)
{
   float retVal;
   char *floatToConvert = (char*) &inFloat;
   char *returnFloat    = (char*) &retVal;

   // swap the bytes into a temporary buffer
   returnFloat[0] = floatToConvert[3];
   returnFloat[1] = floatToConvert[2];
   returnFloat[2] = floatToConvert[1];
   returnFloat[3] = floatToConvert[0];

   return retVal;
}



template <class T> TensorVolume<T>::TensorVolume()
{
	fields[0] = &confidence;
	fields[1] = &dxx;
	fields[2] = &dxy;
	fields[3] = &dxz;
	fields[4] = &dyy;
	fields[5] = &dyz;
	fields[6] = &dzz;
	
	namedFields.insert(make_pair("confidence",fields[0]));
	namedFields.insert(make_pair("dxx",fields[1]));
	namedFields.insert(make_pair("dxy",fields[2]));
	namedFields.insert(make_pair("dxz",fields[3]));
	namedFields.insert(make_pair("dyy",fields[4]));
	namedFields.insert(make_pair("dyz",fields[5]));
	namedFields.insert(make_pair("dzz",fields[6]));
}


template <class T>  void TensorVolume<T>::makeVolume(int width, int height, int depth)
{
	sizeX = width;
	sizeY = height;
	sizeZ = depth;
	
	for(int i=0;i<7;++i)
	   fields[i]->makeVolume(sizeX,sizeY,sizeZ);
}	   

template <class T>  Volume<T>* TensorVolume<T>::component(const char* name)
{
	typename map<string,Volume<T>* >::iterator it = namedFields.find(name);
	return (it==namedFields.end())? 0 : it->second;
}

template <class T> bool TensorVolume<T>::readNRRD(const char* headerfile)
{
	const char* hdot = strrchr(headerfile,'.');
	if (!hdot) 
	{
		cout<<"Error: header file '"<<headerfile<<"' should have extension .nhdr, aborting"<<endl;
		return false;
	}
	
	char data_file[255];								//1. Read NRRD header file to determine various parameters
	bool big_endian;
	if (!readHeader(headerfile,data_file,big_endian)) return false;	
	
	makeVolume(sizeX,sizeY,sizeZ);						//2. Allocate the tensor volumes
	
	if (!readData(data_file,big_endian)) return false;	//3. Read the raw data into the allocated volumes

	return true;
}



template <class T> bool TensorVolume<T>::readHeader(const char* headerfile,char* data_file,bool& big_endian)
{
	FILE* fp = fopen(headerfile,"r");
	if (!fp)
	{
		cout<<"Cannot open header '"<<headerfile<<"', aborting"<<endl;
		fclose(fp);
		return false;
	}
	
	char buf[1024],keyword[32],encoding[32];
	int  sizeAttrs;
	bool first = true;
	while(fgets(buf,1024,fp))		
	{
		if (first)
		{
			if (strncmp(buf,"NRRD00",strlen("NRRD00")))
			{
				cout<<"Error: file does not start with keyword NRRD00xx aborting"<<endl;
				fclose(fp);
				return false;
			}
			else { first=false; continue; }
		}	
			

		if (buf[0]=='#') continue;
		
		
		char* col = strchr(buf,':');
		if (!col)
		{
			cout<<"Error: line "<<buf<<" does not start with a keyword, aborting"<<endl;
			fclose(fp);
			return false;
		}
		
		strncpy(keyword,buf,col-buf);
		keyword[col-buf]='\0';
		char* values = col+1;
		
		if (!strcmp(keyword,"data file"))
		{
			sscanf(values,"%s",data_file);
			char hfile[255];
			strcpy(hfile,headerfile);
			const char* header_path = dirname(hfile);			
			const char* dfile = basename(data_file);
			sprintf(data_file,"%s/%s",header_path,dfile);
		}
		else if (!strcmp(keyword,"sizes"))			
		{
			sscanf(values,"%d %d %d %d",&sizeAttrs,&sizeX,&sizeY,&sizeZ);
		}
		else if (!strcmp(keyword,"encoding"))
		{
			sscanf(values,"%s",encoding);
		}
		else if (!strcmp(keyword,"endian"))
		{
			char endian[32];
			sscanf(values,"%s",endian);
			big_endian = !strcmp(endian,"big");
		}	
	}
	
	fclose(fp);											//Done reading header
	
														//Now check that the read values are admissible:
	
	if (strcmp(encoding,"raw"))
	{
		cout<<"Error, supporting only raw encoding; current value: "<<encoding<<endl;
		return false;
	}
	
	return true;
}





template <class T> bool TensorVolume<T>::readData(const char* datafile,bool big_endian)
{
	FILE* fp = fopen(datafile,"rb");
	
	float val[NumComponents];
	
	unsigned int N = sizeX*sizeY*sizeZ;
	
	bool platform_big_endian = is_big_endian();					//Determine if our platform is little or big endian
	
	for(int z=0;z<sizeZ;++z)
	for(int y=0;y<sizeY;++y)
	for(int x=0;x<sizeX;++x)
	{
		int nr = fread(val,sizeof(float),NumComponents,fp);
		if (nr!=NumComponents)
		{
			cout<<"Error: read only "<<nr<<" values, expected "<<NumComponents<<", aborting"<<endl;
			fclose(fp);
			return false;
		}	
		
		for(int i=0;i<NumComponents;++i)
		{
		   if (big_endian != platform_big_endian)				//Swap data if dataset endianness differs from platform's
		      val[i] = ReverseFloat(val[i]);					//(note: input file is always big endian)
		   (*fields[i])(x,y,z) = val[i];						//Assign read data to the corresponding tensor component
		}   
	}	
	
	fclose(fp);
	
	return true;
}


template class TensorVolume<float>;      
template class TensorVolume<byte>;      
template class TensorVolume<char>;
template class TensorVolume<int>;

