#pragma once

#include "Volume.h"
#include <map>
#include <vector>

class UnstructuredGrid3D;




template <class T> class Isosurface			//Constructs an isosurface from a 3D scalar dataset, stored as a Volume.
{											//The isosurface is returned as an UnstructuredGrid3D (triangle mesh).
public:
				    Isosurface();			//Constructor
				   ~Isosurface();			//Destructor
UnstructuredGrid3D* Contour(const Volume<T>& volume,T level,float cellLengthX=1,float cellLengthY=1,float cellLengthZ=1);
											//Create isosurface at given level from given scalar volume
protected:

struct IsoVertex							//Help structure: stores a vertex of the isosurface
{		
	unsigned int	newID;
	float			x,y,z;
};

struct IsoTriangle							//Help structure: stores a triangle of the isosurface
{											//(represented as a triplet of indices of IsoVertex structures
	unsigned int	pointID[3];
};

typedef std::vector<IsoTriangle>		 IsoTriangles;					//All triangles forming the isosurface mesh	
typedef std::map<unsigned int,IsoVertex> IsoVertices;					//All vertices forming the isosurface mesh, hashed by vertex-id

IsoVertices			isoVertices;										// Set of points which form the isosurface.
IsoTriangles		isoTriangles;										// Set of triangles which form the triangulation of the isosurface.
const Volume<T>*	volume;												// The 3D scalar volume to contour
T					isoLevel;											// The isosurface value.
unsigned int		m_nCellsX, m_nCellsY, m_nCellsZ;					// No. of cells in x, y, and z directions.
float				m_fCellLengthX, m_fCellLengthY, m_fCellLengthZ;		// Cell sizes in x, y, and z directions.

unsigned int		GetEdgeID(unsigned int nX,unsigned int nY,unsigned int nZ,unsigned int nEdgeNo);

unsigned int		GetVertexID(unsigned int nX,unsigned int nY,unsigned int nZ);

IsoVertex			CalculateIntersection(unsigned int nX,unsigned int nY,unsigned int nZ,unsigned int nEdgeNo);
					//Intersects an edge starting at vertex (nX,nY,nZ) in given direction with the isosurface
	
IsoVertex			Interpolate(float fX1,float fY1,float fZ1,float fX2,float fY2,float fZ2);
					//Interpolates between two grid points, returns intersection of edge (1,2) with isosurface
 
void				AddVertex(unsigned int x,unsigned int y,unsigned int z,int edgeNo);
					//Compute and add a new vertex to isosurface
 
UnstructuredGrid3D* CreateIsoMesh();
					//Save the computed isosurface into an UnstructuredGrid3D

void				Clear();
					//Clear all internal (temporary) data structures after computing an isosurface	

static const unsigned int edgeTable[256];
static const unsigned int triTable[256][16];
static const int		  edgeIDTable[12][4];
					//Lookup tables used for the isosurface construction
};







template <class T> inline unsigned int Isosurface<T>::GetVertexID(unsigned int nX,unsigned int nY,unsigned int nZ)
{
	return 3*(nZ*(m_nCellsY + 1)*(m_nCellsX + 1) + nY*(m_nCellsX + 1) + nX);
}


template <class T> inline unsigned int Isosurface<T>::GetEdgeID(unsigned int nX,unsigned int nY,unsigned int nZ,unsigned int nEdgeNo)
{
	if (nEdgeNo>11) return -1;
	const int* eit = edgeIDTable[nEdgeNo];	
	return GetVertexID(nX+eit[0],nY+eit[1],nZ+eit[2])+eit[3];
}

